import pytest

from npe2.io_utils import read, read_get_reader, write
from npe2.types import FullLayerData


def test_read(uses_sample_plugin):
    assert read("some.fzzy") == [(None,)]


def test_read_with_plugin(uses_sample_plugin):
    # no such plugin name.... but skips over the sample plugin
    with pytest.raises(ValueError):
        read("some.fzzy", plugin_name="nope")


def test_read_return_reader(uses_sample_plugin):
    data, reader = read_get_reader("some.fzzy")
    assert data == [(None,)]
    assert reader.command == "my_plugin.some_reader"


null_image: FullLayerData = ([], {}, "image")


def test_writer_exec(uses_sample_plugin):
    # the sample writer knows how to handle two image layers
    result = write("test.tif", [null_image, null_image])
    assert result == ["test.tif"]


@pytest.mark.parametrize("layer_data", [[null_image, null_image], []])
def test_writer_exec_fails(layer_data, uses_sample_plugin):
    # the sample writer doesn't accept no extension
    with pytest.raises(ValueError):
        write("test/path", layer_data)


def test_writer_exec_fails2(uses_sample_plugin):
    # the sample writer doesn't accept 5 images
    with pytest.raises(ValueError):
        write("test.tif", [null_image, null_image, null_image, null_image, null_image])


def test_writer_single_layer_api_exec(uses_sample_plugin):
    # This writer doesn't do anything but type check.
    paths = write("test/path", [([], {}, "labels")])
    assert len(paths) == 1



from typing import List, Optional

from pydantic import BaseModel

from .commands import CommandContribution

# from .configuration import JsonSchemaObject
from .io import ReaderContribution, WriterContribution

# from .keybindings import KeyBindingContribution
from .menus import MenusContribution
from .sample_data import SampleDataContribution
from .submenu import SubmenuContribution
from .themes import ThemeContribution
from .widgets import WidgetContribution


class ContributionPoints(BaseModel):
    commands: Optional[List[CommandContribution]]
    themes: Optional[List[ThemeContribution]]
    readers: Optional[List[ReaderContribution]]
    writers: Optional[List[WriterContribution]]
    sample_data: Optional[List[SampleDataContribution]]
    widgets: Optional[List[WidgetContribution]]

    menus: Optional[MenusContribution]
    submenus: Optional[List[SubmenuContribution]]

    # configuration: Optional[JsonSchemaObject]
    # keybindings: Optional[List[KeyBindingContribution]]



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional

from pydantic import BaseModel, Extra, Field

from ..types import Widget
from .utils import Executable

if TYPE_CHECKING:
    from .._command_registry import CommandRegistry


class WidgetContribution(BaseModel, Executable[Widget]):
    command: str = Field(
        ..., description="Identifier of a command that returns a Widget instance."
    )
    display_name: str = Field(
        ..., description="Name of the widget that will be presented in the display."
    )
    autogenerate: bool = Field(
        default=False,
        description="If true, a widget will be autogenerated from "
        "the type signature of the provided command with magicgui.",
    )

    class Config:
        extra = Extra.forbid

    def get_callable(
        self, _registry: Optional[CommandRegistry] = None
    ) -> Callable[..., Widget]:
        func = super().get_callable()
        if self.autogenerate:
            from magicgui import magic_factory

            return magic_factory(func)
        return func



from typing import Optional

from pydantic import BaseModel
from pydantic.fields import Field

from .utils import Executable


# the existence of the command is not validated at registration-time,
# but rather at call time... (since commands from other extensions can be called)
class KeyBindingContribution(BaseModel, Executable):
    command: str = Field(
        description="Identifier of the command to run when keybinding is triggered."
    )
    key: str = Field(
        description="Key or key sequence (separate keys with plus-sign and sequences "
        "with space, e.g. Ctrl+O and Ctrl+L L for a chord)."
    )
    mac: Optional[str] = Field(description="Mac specific key or key sequence.")
    linux: Optional[str] = Field(description="Linux specific key or key sequence.")
    win: Optional[str] = Field(description="Windows specific key or key sequence.")
    when: Optional[str] = Field(description="Condition when the key is active.")



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"
__author__ = "Talley Lambert"
__email__ = "talley.lambert@gmail.com"

from ._plugin_manager import PluginContext, PluginManager
from .io_utils import read, write
from .manifest import PluginManifest

__all__ = [
    "PluginManifest",
    "PluginManager",
    "PluginContext",
    "write",
    "read",
]



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Tuple, Union, overload

from typing_extensions import Literal

from . import PluginManager
from .types import FullLayerData, LayerData, PathLike

if TYPE_CHECKING:
    from .manifest.io import ReaderContribution, WriterContribution


def read(path: PathLike, *, plugin_name: Optional[str] = None) -> List[LayerData]:
    """Try to read file at `path`, with plugins offering a ReaderContribution.

    Parameters
    ----------
    path : str or Path
        Path to the file or resource being read.
    plugin_name : str, optional
        Optional plugin name.  If provided, only readers from this plugin will be
        tried (it's possible that none will be compatible). by default None

    Returns
    -------
    List[LayerData]
        If a compatible reader is found and executed, a list of LayerDataTuples is
        returned

    Raises
    ------
    ValueError
        If no readers are found or none return data
    """
    return _read(path, plugin_name=plugin_name)


def read_get_reader(
    path: PathLike, *, plugin_name: Optional[str] = None
) -> Tuple[List[LayerData], ReaderContribution]:
    """Variant of `read` that also returns the `ReaderContribution` used."""
    return _read(path, plugin_name=plugin_name, return_reader=True)


def write(
    path: str,
    layer_data: List[FullLayerData],
    *,
    plugin_name: Optional[str] = None,
) -> List[str]:
    """Write layer_data tuples to `path`.

    Parameters
    ----------
    path : str
        The path (file, directory, url) to write.
    layer_data : list of layer data tuples
        List of tuples in the form (data, metadata_dict, layer_type_string)
    plugin_name : str, optional
        Name of the plugin to write data with. If `None` then all plugins
        corresponding to appropriate hook specification will be looped
        through to find the first one that can write the data.

    Returns
    -------
    list of str
        List of file paths that were written

    Raises
    ------
    ValueError
        If no suitable writers are found.
    """
    return _write(path, layer_data, plugin_name=plugin_name)


def write_get_writer(
    path: str,
    layer_data: List[FullLayerData],
    *,
    plugin_name: Optional[str] = None,
) -> Tuple[List[str], WriterContribution]:
    """Variant of write that also returns the `WriterContribution` used."""
    return _write(path, layer_data, plugin_name=plugin_name, return_writer=True)


# -----------------------------------------------------------------------------------


@overload
def _read(
    path: PathLike,
    *,
    plugin_name: Optional[str] = None,
    return_reader: Literal[False] = False,
    _pm=None,
) -> List[LayerData]:
    ...


@overload
def _read(
    path: PathLike,
    *,
    plugin_name: Optional[str] = None,
    return_reader: Literal[True],
    _pm=None,
) -> Tuple[List[LayerData], ReaderContribution]:
    ...


def _read(
    path: PathLike,
    *,
    plugin_name: Optional[str] = None,
    return_reader: bool = False,
    _pm: Optional[PluginManager] = None,
) -> Union[Tuple[List[LayerData], ReaderContribution], List[LayerData]]:
    """Execute the `read...` functions above."""
    if _pm is None:
        _pm = PluginManager.instance()

    for rdr in _pm.iter_compatible_readers(path):
        if plugin_name and not rdr.command.startswith(plugin_name):
            continue
        read_func = rdr.exec(kwargs={"path": path})
        if read_func is not None:
            # if the reader function raises an exception here, we don't try to catch it
            layer_data = read_func(path)
            if layer_data:
                return (layer_data, rdr) if return_reader else layer_data
    raise ValueError(f"No readers returned data for {path!r}")


@overload
def _write(
    path: str,
    layer_data: List[FullLayerData],
    *,
    plugin_name: Optional[str] = None,
    return_writer: Literal[False] = False,
    _pm: Optional[PluginManager] = None,
) -> List[str]:
    ...


@overload
def _write(
    path: str,
    layer_data: List[FullLayerData],
    *,
    plugin_name: Optional[str] = None,
    return_writer: Literal[True],
    _pm: Optional[PluginManager] = None,
) -> Tuple[List[str], WriterContribution]:
    ...


def _write(
    path: str,
    layer_data: List[FullLayerData],
    *,
    plugin_name: Optional[str] = None,
    return_writer: bool = False,
    _pm: Optional[PluginManager] = None,
) -> Union[List[str], Tuple[List[str], WriterContribution]]:

    if not layer_data:
        raise ValueError("Must provide layer data")
    if _pm is None:
        _pm = PluginManager.instance()

    layer_types = [x[2] for x in layer_data]
    writer, new_path = _pm.get_writer(
        path, layer_types=layer_types, plugin_name=plugin_name
    )

    if not writer:
        raise ValueError(f"No writer found for {path!r} with layer types {layer_types}")

    # Writers that take at most one layer must use the single-layer api.
    # Otherwise, they must use the multi-layer api.
    n = sum(ltc.max() for ltc in writer.layer_type_constraints())
    args = (new_path, *layer_data[0][:2]) if n <= 1 else (new_path, layer_data)
    res = writer.exec(args=args)

    # napari_get_writer-style writers don't always return a list
    # though strictly speaking they should?
    result = [res] if isinstance(res, str) else res or []  # type: ignore
    return (result, writer) if return_writer else result



from __future__ import annotations

import json
import re
import sys
from contextlib import contextmanager
from enum import Enum
from importlib import import_module, util
from logging import getLogger
from pathlib import Path
from textwrap import dedent
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterator,
    NamedTuple,
    Optional,
    Sequence,
    Union,
)

import pytomlpp as toml
import yaml
from pydantic import BaseModel, Extra, Field, ValidationError, root_validator, validator

from .contributions import ContributionPoints
from .package_metadata import PackageMetadata
from .utils import Version

try:
    from importlib.metadata import Distribution, distributions
except ImportError:
    from importlib_metadata import Distribution, distributions  # type: ignore

if TYPE_CHECKING:
    from importlib.metadata import EntryPoint

spdx_ids = (Path(__file__).parent / "spdx.txt").read_text().splitlines()
SPDX = Enum("SPDX", {i.replace("-", "_"): i for i in spdx_ids})  # type: ignore

logger = getLogger(__name__)

ENTRY_POINT = "napari.manifest"

# The first release of npe2 defines the first engine version: 0.1.0.
# As the contract around plugins evolve the ENGINE_NUMBER should be
# increased follow SemVer rules. Note that sometimes the version number
# will change even though no npe2 code changes.
ENGINE_VERSION = "0.1.0"
_display_name_pattern = re.compile(r"^[^\W_][\w -~]{1,38}[^\W_]$")


class DiscoverResults(NamedTuple):
    manifest: Optional[PluginManifest]
    entrypoint: Optional[EntryPoint]
    error: Optional[Exception]


class PluginManifest(BaseModel):

    # VS Code uses <publisher>.<name> as a unique ID for the extension
    # should this just be the package name ... not the module name? (yes)
    # do we normalize this? (i.e. underscores / dashes ?) (no)
    # TODO: enforce that this matches the package name

    name: str = Field(
        ...,
        description="The name of the plugin. Should correspond to the python "
        "package name for this plugin.",
    )

    display_name: str = Field(
        "",
        description="User-facing text to display as the name of this plugin",
        # Must be 3-40 characters long, containing printable word characters,
        # and must not begin or end with an underscore, white space, or
        # non-word character.
    )

    # Plugins rely on certain guarantees to interoperate propertly with the
    # plugin engine. These include the manifest specification, conventions
    # around python packaging, command api's, etc. Together these form a
    # "contract". The version of this contract is the "engine version."
    #
    # The first release of npe2 defines the first engine version.
    # As the contract around plugins evolve the ENGINE_NUMBER should be
    # increased follow SemVer rules. Note that sometimes the version number
    # will change even though no npe2 code changes.
    #
    # The `engine` field declares the version of the contract that this plugin
    # targets. Since there is no other version at the moment, it defaults to
    # the current npe2 engine version.
    engine: str = Field(
        ENGINE_VERSION,
        description="A SemVer compatible version string matching the versions "
        "of the plugin engine that the extension is compatible with.",
    )

    # TODO:
    # Perhaps we should version the plugin interface (not so the manifest, but
    # the actual mechanism/consumption of plugin information) independently
    # of napari itself

    # TODO: refactor entry_point to binding points for activate,deactivate
    # TODO: Point to activate function
    # TODO: Point to deactivate function

    # The module that has the activate() function
    entry_point: Optional[str] = Field(
        default=None,
        description="The extension entry point. This should be a fully "
        "qualified module string. e.g. `foo.bar.baz` for a module containing "
        "the plugin's activate() function.",
    )

    contributions: Optional[ContributionPoints]

    _manifest_file: Optional[Path] = None
    package_metadata: Optional[PackageMetadata] = None

    @property
    def license(self) -> Optional[str]:
        return self.package_metadata.license if self.package_metadata else None

    @property
    def version(self) -> Optional[str]:
        return self.package_metadata.version if self.package_metadata else None

    @property
    def description(self) -> Optional[str]:
        return self.package_metadata.summary if self.package_metadata else None

    @property
    def author(self) -> Optional[str]:
        return self.package_metadata.author if self.package_metadata else None

    @root_validator
    def _check_engine_version(cls, values: dict) -> dict:
        declared_version = Version.parse(values.get("engine", ""))
        current_version = Version.parse(ENGINE_VERSION)
        if current_version < declared_version:
            raise ValueError(
                dedent(
                    f"The declared engine version {declared_version} is "
                    "newer than npe2 engine {current_version}. You may need to "
                    "upgrade npe2."
                )
            )
        return values

    @validator("display_name")
    def validate_display_name(cls, v):
        if not _display_name_pattern.match(v):
            raise ValueError(
                f"{v} is not a valid display_name.  The display_name must "
                "be 3-40 characters long, containing printable word characters, "
                "and must not begin or end with an underscore, white space, or "
                "non-word character."
            )
        return v

    @root_validator
    def _validate_root(cls, values: dict) -> dict:
        invalid_commands = []
        if values.get("contributions") is not None:
            for command in values["contributions"].commands or []:
                id_start_actual = command.id.split(".")[0]
                if values["name"] != id_start_actual:
                    invalid_commands.append(command.id)

        if invalid_commands:
            raise ValueError(
                dedent(
                    f"""Commands identifiers must start with the current package name {values['name']!r}
            the following commands where found to break this assumption:
                {invalid_commands}
            """
                )
            )

        return values

    def toml(self, pyproject=False) -> str:
        with engine_in_fields_set(self):
            d = json.loads(self.json(exclude_unset=True))
            if pyproject:
                d = {"tool": {"napari": d}}
            return toml.dumps(d)

    def yaml(self) -> str:
        with engine_in_fields_set(self):
            return yaml.safe_dump(json.loads(self.json(exclude_unset=True)))

    @classmethod
    def from_distribution(cls, name: str) -> PluginManifest:
        """Return PluginManifest given a distribution (package) name.

        Parameters
        ----------
        name : str
            Name of a python distribution installed in the environment.
            Note: this is the package name, not the top level module name,
            (e.g. "scikit-image", not "skimage").

        Returns
        -------
        PluginManifest
            The parsed manifest.

        Raises
        ------
        ValueError
            If the distribution exists, but does not provide a manifest
        PackageNotFoundError
            If there is no distribution found for `name`
        ValidationError
            If the manifest is not valid
        """
        from importlib.metadata import distribution

        dist = distribution(name)  # may raise PackageNotFoundError
        for ep in dist.entry_points:
            if ep.group == ENTRY_POINT:
                return PluginManifest._from_entrypoint(ep, dist)
        raise ValueError(
            "Distribution {name!r} exists but does not provide a napari manifest"
        )

    @classmethod
    def from_file(cls, path: Union[Path, str]) -> PluginManifest:
        """Parse PluginManifest from a specific file.

        Parameters
        ----------
        path : Path or str
            Path to a manifest.  Must have extension {'.json', '.yaml', '.yml', '.toml'}

        Returns
        -------
        PluginManifest
            The parsed manifest.

        Raises
        ------
        FileNotFoundError
            If `path` does not exist.
        ValueError
            If the file extension is not in {'.json', '.yaml', '.yml', '.toml'}
        """
        path = Path(path).expanduser().absolute().resolve()
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")

        loader: Callable
        if path.suffix.lower() == ".json":
            loader = json.load
        elif path.suffix.lower() == ".toml":
            loader = toml.load
        elif path.suffix.lower() in (".yaml", ".yml"):
            loader = yaml.safe_load
        else:
            raise ValueError(f"unrecognized file extension: {path}")

        with open(path) as f:
            data = loader(f) or {}

        if path.name == "pyproject.toml":
            data = data["tool"]["napari"]

        mf = cls(**data)
        mf._manifest_file = path
        return mf

    class Config:
        use_enum_values = True  # only needed for SPDX
        underscore_attrs_are_private = True
        extra = Extra.forbid

    def _call_func_in_plugin_entrypoint(self, funcname: str, args=()) -> None:
        """convenience to call a function in the plugins entry_point, if declared."""
        if not self.entry_point:
            return None
        mod = import_module(self.entry_point)
        func = getattr(mod, funcname, None)
        if callable(func):
            return func(*args)

    @classmethod
    def discover(
        cls,
        entry_point_group: str = ENTRY_POINT,
        paths: Sequence[Union[str, Path]] = (),
    ) -> Iterator[DiscoverResults]:
        """Discover manifests in the environment.

        This function searches for installed python packages with a matching
        entry point group and then attempts to resolve the manifest file.

        The manifest file should be specified in the plugin's `setup.cfg` or
        `setup.py` file using the [entry point group][1]: "napari.manifest".
        For example, this would be the section for a plugin "npe-tester" with
        "napari.yaml" as the manifest file:

        ```cfg
        [options.entry_points]
        napari.manifest =
            npe2-tester = npe2_tester:napari.yaml
        ```

        The manifest file is specified relative to the submodule root path.
        So for the example it will be loaded from:
        `<path/to/npe2-tester>/napari.yaml`.

        [1]: https://packaging.python.org/specifications/entry-points/

        Parameters
        ----------
        entry_point_group : str, optional
            name of entry point group to discover, by default 'napari.manifest'
        paths : Sequence[str], optional
            paths to add to sys.path while discovering.

        Yields
        ------
        DiscoverResults: (3 namedtuples: manifest, entrypoint, error)
            3-tuples with either manifest or (entrypoint and error) being None.
        """
        with temporary_path_additions(paths):
            for dist in distributions():
                for ep in dist.entry_points:
                    if ep.group != entry_point_group:
                        continue
                    try:
                        pm = cls._from_entrypoint(ep, dist)
                        yield DiscoverResults(pm, ep, None)
                    except ValidationError as e:
                        logger.warning(msg=f"Invalid schema {ep.value!r}")
                        yield DiscoverResults(None, ep, e)
                    except Exception as e:
                        logger.error(
                            "%s -> %r could not be imported: %s"
                            % (entry_point_group, ep.value, e)
                        )
                        yield DiscoverResults(None, ep, e)

    @classmethod
    def _from_entrypoint(
        cls, entry_point: EntryPoint, distribution: Optional[Distribution] = None
    ) -> PluginManifest:

        match = entry_point.pattern.match(entry_point.value)  # type: ignore
        module = match.group("module")

        spec = util.find_spec(module or "")
        if not spec:
            raise ValueError(
                f"Cannot find module {module!r} declared in "
                f"entrypoint: {entry_point.value!r}"
            )

        match = entry_point.pattern.match(entry_point.value)  # type: ignore
        fname = match.group("attr")

        for loc in spec.submodule_search_locations or []:
            mf_file = Path(loc) / fname
            if mf_file.exists():
                mf = PluginManifest.from_file(mf_file)
                if distribution is not None:
                    meta = PackageMetadata.from_dist_metadata(distribution.metadata)
                    mf.package_metadata = meta

                    assert mf.name == meta.name, "Manifest name must match package name"
                    return mf

        raise FileNotFoundError(f"Could not find file {fname!r} in module {module!r}")

    @classmethod
    def _from_package_or_name(
        cls, package_or_filename: Union[Path, str]
    ) -> PluginManifest:
        """Internal convenience function, calls both `from_file` and `from_distribution`

        Parameters
        ----------
        package_or_filename : Union[Path, str]
            Either a filename or a package name.  Will be tried first as a filename, and
            then as a distribution name.

        Returns
        -------
        PluginManifest
            [description]

        Raises
        ------
        ValidationError
            If the name can be resolved as either a distribution name or a file,
            but the manifest is not valid.
        ValueError
            If the name does not resolve to either a distribution name or a filename.

        """
        from pydantic import ValidationError

        from npe2 import PluginManifest

        try:
            return PluginManifest.from_file(package_or_filename)
        except ValidationError:
            raise
        except (FileNotFoundError, ValueError):
            try:
                return PluginManifest.from_distribution(str(package_or_filename))
            except ValidationError:
                raise
            except Exception as e:
                raise ValueError(
                    f"Could not find manifest for {package_or_filename!r} as either a "
                    "package name or a file.."
                ) from e

    ValidationError = ValidationError  # for convenience of access


@contextmanager
def temporary_path_additions(paths: Sequence[Union[str, Path]] = ()):
    for p in reversed(paths):
        sys.path.insert(0, str(p))
    try:
        yield
    finally:
        for p in paths:
            sys.path.remove(str(p))


@contextmanager
def engine_in_fields_set(manifest: PluginManifest):
    was_there = "engine" in manifest.__fields_set__
    manifest.__fields_set__.add("engine")
    try:
        yield
    finally:
        if not was_there:
            manifest.__fields_set__.discard("engine")


if __name__ == "__main__":
    print(PluginManifest.schema_json())



import sys
from pathlib import Path
from unittest.mock import patch

import pytest

from npe2 import PluginManager, PluginManifest


@pytest.fixture
def sample_path():
    return Path(__file__).parent / "sample"


@pytest.fixture
def uses_sample_plugin(sample_path):
    sys.path.append(str(sample_path))
    pm = PluginManager.instance()
    pm.discover()
    yield
    sys.path.remove(str(sample_path))


@pytest.fixture
def plugin_manager():
    pm = PluginManager()
    pm.discover()
    return pm


@pytest.fixture(autouse=True)
def mock_discover():
    _discover = PluginManifest.discover

    def wrapper(*args, **kwargs):
        before = sys.path.copy()
        # only allowing things from test directory in discover
        sys.path = [x for x in sys.path if str(Path(__file__).parent) in x]
        try:
            yield from _discover(*args, **kwargs)
        finally:
            sys.path = before

    with patch("npe2.PluginManifest.discover", wraps=wrapper):
        yield



from typing import Optional, Union

from pydantic import BaseModel, color
from pydantic.fields import Field
from typing_extensions import Literal


# pydantic doesn't implement color equality?
class Color(color.Color):
    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, color.Color):
            return False
        return self.as_rgb_tuple() == __o.as_rgb_tuple()


class ThemeColors(BaseModel):
    canvas: Optional[Color]
    console: Optional[Color]
    background: Optional[Color]
    foreground: Optional[Color]
    primary: Optional[Color]
    secondary: Optional[Color]
    highlight: Optional[Color]
    text: Optional[Color]
    icon: Optional[Color]
    warning: Optional[Color]
    current: Optional[Color]


class ThemeContribution(BaseModel):
    label: str = Field(description="Label of the color theme as shown in the UI.")
    id: str = Field(description="Id of the color theme as used in the user settings.")
    type: Union[Literal["dark"], Literal["light"]] = Field(
        description="Base theme type, used for icons and filling in unprovided colors"
    )
    colors: ThemeColors = Field(description="Theme colors")



import json
import sys

import pytest
from pydantic import ValidationError

from npe2 import PluginManager, PluginManifest
from npe2.cli import main
from npe2.manifest.package_metadata import PackageMetadata


def test_schema():
    assert isinstance(PluginManifest.schema_json(), str)

    dschema = PluginManifest.schema()
    assert isinstance(dschema, dict)
    assert "name" in dschema["properties"]


def test_discover_empty():
    # sanity check to make sure sample_plugin must be in path
    manifests = [
        result.manifest.name for result in PluginManifest.discover() if result.manifest
    ]

    assert "my_plugin" not in manifests


def test_sample_plugin_valid(sample_path):
    assert PluginManifest.from_file(sample_path / "my_plugin" / "napari.yaml")


def test_discover(uses_sample_plugin):
    discover_results = list(PluginManifest.discover())
    assert len(discover_results) == 1
    [(manifest, entrypoint, error)] = discover_results
    assert manifest and manifest.name == "my_plugin"
    assert entrypoint and entrypoint.group == "napari.manifest"
    assert entrypoint.value == "my_plugin:napari.yaml"
    assert error is None


def test_package_meta(uses_sample_plugin):
    direct_meta = PackageMetadata.for_package("my_plugin")
    assert direct_meta.name == "my_plugin"
    assert direct_meta.version == "1.2.3"
    discover_results = list(PluginManifest.discover())
    [(manifest, *_)] = discover_results
    assert manifest
    assert manifest.package_metadata == direct_meta


def test_all_package_meta():
    """make sure PackageMetadata works for whatever packages are in the environment.

    just a brute force way to get a little more validation coverage
    """
    try:
        from importlib.metadata import distributions
    except ImportError:
        from importlib_metadata import distributions  # type: ignore

    for d in distributions():
        assert PackageMetadata.from_dist_metadata(d.metadata)


def test_cli(monkeypatch, sample_path):
    cmd = ["npe2", "validate", str(sample_path / "my_plugin" / "napari.yaml")]
    monkeypatch.setattr(sys, "argv", cmd)
    with pytest.raises(SystemExit) as e:
        main()
    assert e.value.code == 0


def _mutator_1(data):
    data["name"] = "invalid??"
    return data


def _mutator_2(data):
    data["name"] = "dash-invalid"
    return data


def _mutator_3(data):
    assert "contributions" in data
    c = data["contributions"]["commands"][0]["id"]
    data["contributions"]["commands"][0]["id"] = ".".join(
        ["not_packagename", *c.split(".")[1:]]
    )
    return data


def _mutator_4(data):
    assert "contributions" in data
    data["contributions"]["commands"][0]["python_name"] = "this.has.no.colon"
    return data


def _mutator_5(data):
    assert "contributions" in data
    data["contributions"]["commands"][0]["python_name"] = "1starts_with_number"
    return data


def _valid_mutator_no_contributions(data):
    """
    Contributions can be absent, in which case the Pydantic model will set the
    default value to None, and not the empty list, make sure that works.
    """
    del data["contributions"]
    return data


def _valid_mutator_no_contributions_None(data):
    """
    Contributions can be absent, in which case the Pydantic model will set the
    default value to None, and not the empty list, make sure that works.
    """
    data["contributions"] = None
    return data


def _mutator_no_contributes_extra_field(data):
    """
    Contributions used to be called contributes.

    Check that an extra field fails.
    """
    data["invalid_extra_name"] = data["contributions"]
    del data["contributions"]
    return data


def _mutator_writer_requires_non_empty_layer_types(data):
    data["contributions"]["writers"][0]["layer_types"] = []
    return data


def _mutator_writer_invalid_layer_type_constraint(data):
    data["contributions"]["writers"][0]["layer_types"].append("image{")
    return data


def _mutator_writer_invalid_file_extension_1(data):
    data["contributions"]["writers"][0]["filename_extensions"] = ["*"]
    return data


def _mutator_writer_invalid_file_extension_2(data):
    data["contributions"]["writers"][0]["filename_extensions"] = ["."]
    return data


def _mutator_engine_version_too_high(data):
    data["engine"] = "999.999.999"
    return data


@pytest.mark.parametrize(
    "mutator",
    [
        _mutator_1,
        _mutator_2,
        _mutator_3,
        _mutator_4,
        _mutator_5,
        _mutator_no_contributes_extra_field,
        _mutator_writer_requires_non_empty_layer_types,
        _mutator_writer_invalid_layer_type_constraint,
        _mutator_writer_invalid_file_extension_1,
        _mutator_writer_invalid_file_extension_2,
        _mutator_engine_version_too_high,
    ],
)
def test_invalid(mutator, uses_sample_plugin):
    result = next(
        result
        for result in PluginManifest.discover()
        if result.manifest and result.manifest.name == "my_plugin"
    )
    assert result.error is None
    assert result.manifest is not None
    pm = result.manifest
    data = json.loads(pm.json(exclude_unset=True))
    mutator(data)
    with pytest.raises(ValidationError):
        PluginManifest(**data)


@pytest.mark.parametrize(
    "mutator",
    [_valid_mutator_no_contributions, _valid_mutator_no_contributions_None],
)
def test_valid_mutations(mutator, uses_sample_plugin):

    assert mutator.__name__.startswith("_valid")

    pm = list(PluginManifest.discover())[0]
    assert pm.manifest
    # make sure the data is a copy as we'll mutate it
    data = json.loads(pm.manifest.json(exclude_unset=True))
    mutator(data)

    PluginManifest(**data)


@pytest.mark.parametrize(
    "display_name",
    [
        "Here there everywhere and more with giggles and friends",
        "ab",
        " abc",
        "abc ",
        "_abc",
        "abc_",
        "abc♱",
    ],
)
def test_invalid_display_names(display_name, uses_sample_plugin):
    field = PluginManifest.__fields__["display_name"]
    value, err = field.validate(display_name, {}, loc="display_name")
    assert err is not None


@pytest.mark.parametrize(
    "display_name",
    [
        "Some Cell & Stru买cture Segmenter",
        "Segment Blobs and Things with Membranes",
        "abc",
        "abc䜁䜂",
    ],
)
def test_valid_display_names(display_name, uses_sample_plugin):
    field = PluginManifest.__fields__["display_name"]
    value, err = field.validate(display_name, {}, loc="display_name")
    assert err is None


def test_display_name_default_is_valid():
    PluginManifest(name="", entry_point="")


def test_writer_empty_layers():
    pm = PluginManager()
    pm.discover()
    writers = list(pm.iter_compatible_writers([]))
    assert not writers


@pytest.mark.parametrize(
    "param",
    [
        (["image"] * 2, 1),
        (["labels"], 0),
        (["image"] * 4, 1),
        (["image"] * 5, 0),
        (["points", "surface"], 1),
        (["points", "surface", "points"], 0),
    ],
)
def test_writer_ranges(param, uses_sample_plugin, plugin_manager: PluginManager):
    layer_types, expected_count = param
    nwriters = sum(
        w.command == "my_plugin.my_writer"
        for w in plugin_manager.iter_compatible_writers(layer_types)
    )

    assert nwriters == expected_count


@pytest.mark.parametrize(
    "expr",
    [
        "vectors{",
        "image",  # should parse fine, but be a duplication error
        "vectors{8,3}",
        "vectors{-1}",
        "vectors??",
        "other?",
    ],
)
def test_writer_invalid_layer_type_expressions(expr, uses_sample_plugin):
    result = next(
        result
        for result in PluginManifest.discover()
        if result.manifest and result.manifest.name == "my_plugin"
    )
    assert result.error is None
    assert result.manifest is not None
    pm = result.manifest
    data = json.loads(pm.json(exclude_unset=True))

    assert "contributions" in data
    assert "writers" in data["contributions"]
    data["contributions"]["writers"][0]["layer_types"].append(expr)

    with pytest.raises(ValidationError):
        PluginManifest(**data)


@pytest.mark.parametrize(
    "expr",
    ["vectors", "vectors+", "vectors*", "vectors?", "vectors{3}", "vectors{3,8}"],
)
def test_writer_valid_layer_type_expressions(expr, uses_sample_plugin):
    result = next(
        result
        for result in PluginManifest.discover()
        if result.manifest and result.manifest.name == "my_plugin"
    )
    assert result.error is None
    assert result.manifest is not None
    pm = result.manifest
    data = json.loads(pm.json(exclude_unset=True))

    assert "contributions" in data
    assert "writers" in data["contributions"]
    data["contributions"]["writers"][0]["layer_types"].append(expr)

    PluginManifest(**data)


def test_widgets(uses_sample_plugin, plugin_manager: PluginManager):
    from magicgui._magicgui import MagicFactory

    widgets = list(plugin_manager.iter_widgets())
    assert len(widgets) == 2
    assert widgets[0].command == "my_plugin.some_widget"
    w = widgets[0].exec()
    assert type(w).__name__ == "SomeWidget"

    assert widgets[1].command == "my_plugin.some_function_widget"
    w = widgets[1].get_callable()
    assert isinstance(w, MagicFactory)


def test_sample(uses_sample_plugin, plugin_manager: PluginManager):
    plugin, contribs = list(plugin_manager.iter_sample_data())[0]
    assert plugin == "my_plugin"
    assert len(contribs) == 2
    ctrbA, ctrbB = contribs
    # ignoring types because .command and .uri come from different sample provider
    # types... they don't both have "command" or "uri"
    assert ctrbA.command == "my_plugin.generate_random_data"
    assert ctrbB.uri == "https://picsum.photos/1024"
    assert isinstance(ctrbA.open(), list)
    assert isinstance(ctrbB.open(), list)


def test_toml_round_trip(sample_path, tmp_path):
    pm = PluginManifest.from_file(sample_path / "my_plugin" / "napari.yaml")

    toml_file = tmp_path / "napari.toml"
    toml_file.write_text(pm.toml())

    pm2 = PluginManifest.from_file(toml_file)
    assert pm == pm2


def test_json_round_trip(sample_path, tmp_path):
    pm = PluginManifest.from_file(sample_path / "my_plugin" / "napari.yaml")

    json_file = tmp_path / "napari.json"
    json_file.write_text(pm.json())

    pm2 = PluginManifest.from_file(json_file)
    assert pm == pm2



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, Union

from pydantic import BaseModel
from pydantic.fields import Field

from ..types import LayerData
from .utils import Executable

if TYPE_CHECKING:
    from .._command_registry import CommandRegistry


class _SampleDataContribution(BaseModel, ABC):
    display_name: str = Field(
        ..., description="String to show in the GUI when referring to this sample"
    )
    key: str = Field(
        ...,
        description="A key to identify this sample. Must be unique across the various "
        "samples provided by a single plugin",
    )

    @abstractmethod
    def open(
        self, *args, _registry: Optional["CommandRegistry"] = None, **kwargs
    ) -> List[LayerData]:
        ...


class SampleDataGenerator(_SampleDataContribution, Executable[List[LayerData]]):
    command: str = Field(
        ..., description="Identifier of a command that returns layer data tuple."
    )

    def open(
        self, *args, _registry: Optional["CommandRegistry"] = None, **kwargs
    ) -> List[LayerData]:
        return self.exec(args, kwargs, _registry=_registry)


class SampleDataURI(_SampleDataContribution):
    uri: str = Field(
        ...,
        description="Path or URL to a data resource. "
        "This URI should be a valid input to `io_utils.read`",
    )
    reader_plugin: Optional[str] = Field(
        None,
        description="Name of plugin to use to open URI",
    )

    def open(self, *args, **kwargs) -> List[LayerData]:
        from ..io_utils import read

        return read(self.uri, plugin_name=self.reader_plugin)


SampleDataContribution = Union[SampleDataGenerator, SampleDataURI]



import re
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel, Extra, Field, validator

if TYPE_CHECKING:
    from .._command_registry import CommandRegistry

_distname = "([a-zA-Z_][a-zA-Z0-9_-]+)"
_identifier = "([a-zA-Z_][a-zA-Z_0-9]+)"
_identifier_plus_dash = "([a-zA-Z_][a-zA-Z_0-9-]+)"

# how do we deal with keywords ?
# do we try to validate ? Or do we just
# assume users won't try to create a command named
# `npe2_tester.False.if.for.in` ?
_dotted_name = f"(({_identifier_plus_dash}\\.)*{_identifier_plus_dash})"
_python_name_pattern = re.compile(f"^{_dotted_name}:{_dotted_name}$")


class CommandContribution(BaseModel):
    """Contribute a command.

    Contribute the UI for a command consisting of a title and (optionally) an
    icon, category, and enabled state. Enablement is expressed with when
    clauses. By default, commands show in the Command Palette (⇧⌘P) but they
    can also show in other menus.

    Presentation of contributed commands depends on the containing menu. The
    Command Palette, for instance, prefixes commands with their category,
    allowing for easy grouping. However, the Command Palette doesn't show icons
    nor disabled commands. The editor context menu, on the other hand, shows
    disabled items but doesn't show the category label.

    Note: When a command is invoked (from a key binding, from the Command
    Palette, any other menu, or programmatically), VS Code will emit an
    activationEvent onCommand:${command}.
    """

    id: str = Field(
        ...,
        description=dedent(
            """
        Identifier of the command to execute

        While this may looks a python fully qualified name this does not refer
        to a python object.
        This identifier is specific to Napari, and will be considered unique.
        It follow the same rule as Python fully qualified name, with the extra
        restriction as being limited to ascii"""
        ),
        regex=f"^(({_distname}\\.)*{_identifier})$",
    )

    title: str = Field(
        ...,
        description="User facing title representing the command. Example: "
        "'Apply gaussian blur' or 'Open my dock widget.",
    )
    # short_title: Optional[str] = Field(
    #     None,
    #     description="(Optional) Short title by which the command is "
    #     "represented in the UI",
    # )
    # category: Optional[str] = Field(
    #     None,
    #     description="(Optional) Category string by the command is grouped in the UI",
    # )
    # icon: Optional[Union[str, Icon]] = Field(
    #     None,
    #     description=(
    #         "(Optional) Icon which is used to represent the command in the UI."
    #         " Either a file path, an object with file paths for dark and light"
    #         "themes, or a theme icon references, like `$(zap)`"
    #     ),
    # )
    # enablement: Optional[str] = Field(
    #     None,
    #     description=(
    #         "(Optional) Condition which must be true to enable the command in the UI "
    #         "(menu and keybindings). Does not prevent executing the command "
    #         "by other means, like the `executeCommand` api."
    #     ),
    # )

    python_name: Optional[str] = Field(
        None,
        description="(Optional) Fully qualified name to callable python object "
        "implementing this command. This usually takes the form of "
        "`{obj.__module__}:{obj.__qualname__} (e.g. "
        "`my_package.a_module:some_function`). If provided, using `register_command` "
        "in the plugin activate function is optional (but takes precedence).",
    )

    @validator("python_name")
    def validate_python_name(cls, v):
        # test for regex validation.
        if v and not _python_name_pattern.match(v):
            raise ValueError(
                f"{v} is not a valid python_name.  A python_name must "
                "be of the form `{obj.__module__}:{obj.__qualname__} `(e.g. "
                "`my_package.a_module:some_function`). "
            )
        return v

    class Config:
        extra = Extra.forbid

    def exec(
        self,
        args: tuple = (),
        kwargs: dict = {},
        _registry: Optional["CommandRegistry"] = None,
    ) -> Any:
        if _registry is None:
            from .._plugin_manager import PluginManager

            _registry = PluginManager.instance().commands
        return _registry.execute(self.id, args, kwargs)



from __future__ import annotations

import re
from dataclasses import dataclass
from functools import total_ordering
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    SupportsInt,
    Tuple,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from typing_extensions import Protocol

    from .._command_registry import CommandRegistry

    class ProvidesCommand(Protocol):
        command: str

        def get_callable(self, _registry: Optional[CommandRegistry] = None):
            ...


R = TypeVar("R")


# TODO: add ParamSpec when it's supported better by mypy
class Executable(Generic[R]):
    def exec(
        self: ProvidesCommand,
        args: tuple = (),
        kwargs: dict = {},
        _registry: Optional[CommandRegistry] = None,
    ) -> R:
        return self.get_callable(_registry)(*args, **kwargs)

    def get_callable(
        self: ProvidesCommand,
        _registry: Optional[CommandRegistry] = None,
    ) -> Callable[..., R]:
        if _registry is None:
            from .._plugin_manager import PluginManager

            _registry = PluginManager.instance().commands
        return _registry.get(self.command)

    @property
    def plugin_name(self: ProvidesCommand):
        # takes advantage of the fact that command always starts with manifest.name
        return self.command.split(".")[0]


@total_ordering
@dataclass
class Version:
    """A semver compatible version class.

    mostly vendored from python-semver (BSD-3):
    https://github.com/python-semver/python-semver/
    """

    major: SupportsInt
    minor: SupportsInt = 0
    patch: SupportsInt = 0
    prerelease: Union[bytes, str, int, None] = None
    build: Union[bytes, str, int, None] = None

    _SEMVER_PATTERN = re.compile(
        r"""
            ^
            (?P<major>0|[1-9]\d*)
            \.
            (?P<minor>0|[1-9]\d*)
            \.
            (?P<patch>0|[1-9]\d*)
            (?:-(?P<prerelease>
                (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
                (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
            ))?
            (?:\+(?P<build>
                [0-9a-zA-Z-]+
                (?:\.[0-9a-zA-Z-]+)*
            ))?
            $
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, version: Union[bytes, str]) -> Version:
        """Convert string or bytes into Version object."""
        if isinstance(version, bytes):
            version = version.decode("UTF-8")
        match = cls._SEMVER_PATTERN.match(version)
        if match is None:
            raise ValueError(f"{version} is not valid SemVer string")
        matched_version_parts: Dict[str, Any] = match.groupdict()
        return cls(**matched_version_parts)

    # NOTE: we're only comparing the numeric parts for now.
    # ALSO: the rest of the comparators come  from functools.total_ordering
    def __eq__(self, other) -> bool:
        return self.to_tuple()[:3] == self._from_obj(other).to_tuple()[:3]

    def __lt__(self, other) -> bool:
        return self.to_tuple()[:3] < self._from_obj(other).to_tuple()[:3]

    @classmethod
    def _from_obj(cls, other):
        if isinstance(other, (str, bytes)):
            other = Version.parse(other)
        elif isinstance(other, dict):
            other = Version(**other)
        elif isinstance(other, (tuple, list)):
            other = Version(*other)
        elif not isinstance(other, Version):
            raise TypeError(
                f"Expected str, bytes, dict, tuple, list, or {cls} instance, "
                f"but got {type(other)}"
            )
        return other

    def to_tuple(self) -> Tuple[int, int, int, Optional[str], Optional[str]]:
        """Return version as tuple (first three are int, last two Opt[str])."""
        return (
            int(self.major),
            int(self.minor),
            int(self.patch),
            str(self.prerelease) if self.prerelease is not None else None,
            str(self.build) if self.build is not None else None,
        )

    def __iter__(self):
        yield from self.to_tuple()



import warnings
from pathlib import Path
from textwrap import indent
from typing import Optional

import typer

from npe2 import PluginManifest

app = typer.Typer()


def _pprint_yaml(string):
    """Print yaml nicely, depending on available modules."""
    try:
        from rich.console import Console
        from rich.syntax import Syntax

        Console().print(Syntax(string, "yaml", theme="fruity"))
    except Exception:
        typer.echo(string)


@app.command()
def validate(name: str, debug: bool = False):
    """Validate manifest for a distribution name or manifest filepath."""

    valid = False
    try:
        pm = PluginManifest._from_package_or_name(name)
        msg = f"✔ Manifest for {(pm.display_name or pm.name)!r} valid!"
        valid = True
    except PluginManifest.ValidationError as err:
        msg = f"🅇 Invalid! {err}"
    except Exception as err:
        msg = f"🅇 Failed to load {name!r}. {type(err).__name__}: {err}"
        if debug:
            raise

    typer.secho(msg, fg=typer.colors.GREEN if valid else typer.colors.RED, bold=True)


@app.command()
def parse(name: str):
    """Show parsed manifest as yaml"""
    pm = PluginManifest._from_package_or_name(name)
    _pprint_yaml(pm.yaml())


@app.command()
def convert(
    path: Path = typer.Argument(
        ...,
        help="Path of a local repository to convert (package must also be installed in"
        " current environment). Or, the name of an installed package/plugin. If a "
        "package is provided instead of a directory, the new manifest will simply be "
        "printed to stdout.",
    ),
    dry_run: Optional[bool] = typer.Option(
        False,
        "--dry-runs",
        "-n",
        help="Just print manifest to stdout. Do not modify anything",
    ),
):
    """Convert first generation napari plugin to new (manifest) format."""
    from ._from_npe1 import convert_repository, manifest_from_npe1

    try:
        with warnings.catch_warnings(record=True) as w:
            if path.is_dir():
                pm, mf_path = convert_repository(path, dry_run=dry_run)
            else:
                pm = manifest_from_npe1(str(path))
        if w:
            typer.secho("Some issues occured:", fg=typer.colors.RED, bold=False)
            for r in w:
                typer.secho(
                    indent(str(r.message), "  "),
                    fg=typer.colors.MAGENTA,
                    bold=False,
                )
            print()
    except Exception as e:
        msg = f"Conversion failed:\n{type(e).__name__}: {e}"
        typer.secho(msg, fg=typer.colors.RED, bold=True)
        raise typer.Exit(1)

    if dry_run or not path.is_dir():
        if path.is_dir():
            typer.secho(
                f"# Manifest would be written to {mf_path}",
                fg=typer.colors.BRIGHT_GREEN,
                bold=False,
            )
        _pprint_yaml(pm.yaml())
    else:
        msg = f"✔  Conversion complete! New manifest at {mf_path}."
        typer.secho(msg, fg=typer.colors.GREEN, bold=True)
        typer.secho(
            "If you have any napari_plugin_engine imports or hook_implementation "
            "decorators, you may remove them now.",
            fg=typer.colors.GREEN,
            bold=False,
        )


def main():
    app()



import pytest
from magicgui import magic_factory
from napari_plugin_engine import napari_hook_implementation

from npe2._from_npe1 import manifest_from_npe1

try:
    from importlib.metadata import PackageNotFoundError
except ImportError:
    from importlib_metadata import PackageNotFoundError  # type: ignore


def gen_data():
    ...


class MyWidget:
    ...


def some_function(x: int):
    ...


some_widget = magic_factory(some_function)


class TestPlugin:
    @napari_hook_implementation
    def napari_get_reader(path):
        ...

    @napari_hook_implementation
    def napari_write_image(path, data, meta):
        ...

    @napari_hook_implementation
    def napari_write_labels(path, data, meta):
        ...

    @staticmethod
    @napari_hook_implementation
    def napari_provide_sample_data():
        return {
            "random data": gen_data,
            "random image": "https://picsum.photos/1024",
            "sample_key": {
                "display_name": "Some Random Data (512 x 512)",
                "data": gen_data,
            },
        }

    @staticmethod
    @napari_hook_implementation
    def napari_experimental_provide_theme():
        return {
            "super_dark": {
                "name": "super_dark",
                "background": "rgb(12, 12, 12)",
                "foreground": "rgb(65, 72, 81)",
                "primary": "rgb(90, 98, 108)",
                "secondary": "rgb(134, 142, 147)",
                "highlight": "rgb(106, 115, 128)",
                "text": "rgb(240, 241, 242)",
                "icon": "rgb(209, 210, 212)",
                "warning": "rgb(153, 18, 31)",
                "current": "rgb(0, 122, 204)",
                "syntax_style": "native",
                "console": "rgb(0, 0, 0)",
                "canvas": "black",
            },
            "pretty_light": {
                "background": "rgb(192, 223, 139)",
            },
        }

    @staticmethod
    @napari_hook_implementation
    def napari_experimental_provide_dock_widget():
        return [MyWidget, (some_widget, {"name": "My Other Widget"})]

    @staticmethod
    @napari_hook_implementation
    def napari_experimental_provide_function():
        return some_function


def test_conversion():
    with pytest.warns(UserWarning):
        assert manifest_from_npe1("svg")


def test_conversion2():
    mf = manifest_from_npe1(module=TestPlugin)
    assert isinstance(mf.dict(), dict)


def test_conversion_missing():
    with pytest.raises(ModuleNotFoundError), pytest.warns(UserWarning):
        manifest_from_npe1("does-not-exist-asdf6as987")


def test_conversion_package_is_not_a_plugin():
    with pytest.raises(PackageNotFoundError):
        manifest_from_npe1("pytest")



import ast
import itertools
import re
import sys
import warnings
from configparser import ConfigParser
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from napari_plugin_engine import (
    HookCaller,
    HookImplementation,
    PluginManager,
    napari_hook_specification,
)

from npe2.manifest import PluginManifest
from npe2.manifest.commands import CommandContribution
from npe2.manifest.themes import ThemeColors
from npe2.manifest.widgets import WidgetContribution

try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata  # type: ignore

NPE1_EP = "napari.plugin"
NPE2_EP = "napari.manifest"


# fmt: off
class HookSpecs:
    def napari_provide_sample_data(): ...  # type: ignore  # noqa: E704
    def napari_get_reader(path): ...  # noqa: E704
    def napari_get_writer(path, layer_types): ...  # noqa: E704
    def napari_write_image(path, data, meta): ...  # noqa: E704
    def napari_write_labels(path, data, meta): ...  # noqa: E704
    def napari_write_points(path, data, meta): ...  # noqa: E704
    def napari_write_shapes(path, data, meta): ...  # noqa: E704
    def napari_write_surface(path, data, meta): ...  # noqa: E704
    def napari_write_vectors(path, data, meta): ...  # noqa: E704
    def napari_experimental_provide_function(): ...  # type: ignore  # noqa: E704
    def napari_experimental_provide_dock_widget(): ...  # type: ignore  # noqa: E704
    def napari_experimental_provide_theme(): ...  # type: ignore  # noqa: E704
# fmt: on


for m in dir(HookSpecs):
    if m.startswith("napari"):
        setattr(HookSpecs, m, napari_hook_specification(getattr(HookSpecs, m)))


@dataclass
class PluginPackage:
    package_name: str
    ep_name: str
    ep_value: str
    top_module: str
    setup_cfg: Optional[Path] = None

    @property
    def name_pairs(self):
        names = (self.ep_name, self.package_name, self.top_module)
        return itertools.product(names, repeat=2)


@lru_cache()
def plugin_packages() -> List[PluginPackage]:
    """List of all packages with napari entry points.

    This is useful to help resolve naming issues (due to the terrible confusion
    around *what* a npe1 plugin name actually was).
    """

    packages = []
    for dist in metadata.distributions():
        for ep in dist.entry_points:
            if ep.group != NPE1_EP:
                continue
            top = dist.read_text("top_level.txt")
            top = top.splitlines()[0] if top else ep.value.split(".")[0]
            packages.append(
                PluginPackage(dist.metadata["Name"], ep.name, ep.value, top)
            )
    return packages


def ensure_package_name(name: str):
    """Try all the tricks we know to find a package name given a plugin name."""
    for attr in ("package_name", "ep_name", "top_module"):
        for p in plugin_packages():
            if name == getattr(p, attr):
                return p.package_name
    raise KeyError(f"Unable to find a locally installed package for plugin {name!r}")


@lru_cache()
def npe1_plugin_manager() -> Tuple[PluginManager, Tuple[int, list]]:
    pm = PluginManager("napari", discover_entry_point=NPE1_EP)
    pm.add_hookspecs(HookSpecs)
    result = pm.discover()
    return pm, result


def norm_plugin_name(plugin_name: Optional[str] = None, module: Any = None) -> str:
    """Try all the things we know to detect something called `plugin_name`."""
    plugin_manager, (_, errors) = npe1_plugin_manager()

    # directly providing a module is mostly for testing.
    if module is not None:
        if plugin_name:  # pragma: no cover
            warnings.warn("module provided, plugin_name ignored")
        plugin_name = getattr(module, "__name__", "dynamic_plugin")
        if not plugin_manager.is_registered(plugin_name):
            plugin_manager.register(module, plugin_name)
        return cast(str, plugin_name)

    if plugin_name in plugin_manager.plugins:
        return cast(str, plugin_name)

    for pkg in plugin_packages():
        for a, b in pkg.name_pairs:
            if plugin_name == a and b in plugin_manager.plugins:
                return b

    # we couldn't find it:
    for e in errors:
        if module and e.plugin == module:
            raise type(e)(e.format())
        for pkg in plugin_packages():
            if plugin_name in (pkg.ep_name, pkg.package_name, pkg.top_module):
                raise type(e)(e.format())

    msg = f"We tried hard! but could not detect a plugin named {plugin_name!r}."
    if plugin_manager.plugins:
        msg += f" Plugins found include: {list(plugin_manager.plugins)}"
    raise metadata.PackageNotFoundError(msg)


def manifest_from_npe1(
    plugin_name: Optional[str] = None, module: Any = None
) -> PluginManifest:
    """Return manifest object given npe1 plugin_name or package name.

    One of `plugin_name` or `module` must be provide.

    Parameters
    ----------
    plugin_name : str
        Name of package/plugin to convert, by default None
    module : Module
        namespace object, to directly import (mostly for testing.), by default None
    """
    plugin_manager, _ = npe1_plugin_manager()
    plugin_name = norm_plugin_name(plugin_name, module)

    _module = plugin_manager.plugins[plugin_name]
    package = ensure_package_name(plugin_name) if module is None else "dynamic"

    parser = HookImplParser(package, plugin_name)
    parser.parse_callers(plugin_manager._plugin2hookcallers[_module])

    return PluginManifest(name=package, contributions=dict(parser.contributions))


class HookImplParser:
    def __init__(self, package: str, plugin_name: str) -> None:
        self.package = package
        self.plugin_name = plugin_name
        self.contributions: DefaultDict[str, list] = DefaultDict(list)

    def parse_callers(self, callers: Iterable[HookCaller]):
        for caller in callers:
            for impl in caller.get_hookimpls():
                if self.plugin_name and impl.plugin_name != self.plugin_name:
                    continue
                # call the corresponding hookimpl parser
                try:
                    getattr(self, impl.specname)(impl)
                except Exception as e:  # pragma: no cover
                    warnings.warn(f"Failed to convert {impl.specname}: {e}")

    def napari_experimental_provide_theme(self, impl: HookImplementation):
        ThemeDict = Dict[str, Union[str, Tuple, List]]
        d: Dict[str, ThemeDict] = impl.function()
        for name, theme_dict in d.items():
            colors = ThemeColors(**theme_dict)
            clr = colors.background or colors.foreground
            luma = _luma(*clr.as_rgb_tuple()[:3]) if clr else 0
            self.contributions["themes"].append(
                {
                    "label": name,
                    "id": name.lower().replace(" ", "_"),
                    "type": "dark" if luma < 128 else "light",
                    "colors": colors,
                }
            )

    def napari_get_reader(self, impl: HookImplementation):
        self.contributions["readers"].append(
            {
                "command": self.add_command(impl),
                "accepts_directories": True,
                "filename_patterns": ["<EDIT_ME>"],
            }
        )

    def napari_provide_sample_data(self, impl: HookImplementation):
        module = sys.modules[impl.function.__module__.split(".", 1)[0]]
        package_dir = module.__file__.rsplit("/", 1)[0]

        samples: Dict[str, Union[dict, str, Callable]] = impl.function()
        for key, sample in samples.items():
            _sample: Union[str, Callable]
            if isinstance(sample, dict):
                display_name = sample.get("display_name")
                _sample = sample.get("data")  # type: ignore
            else:
                _sample = sample
                display_name = key

            _key = _safe_key(key)
            s = {"key": _key, "display_name": display_name}
            if callable(_sample):
                # let these raise exceptions here immediately if they don't validate
                id = f"{self.package}.data.{_key}"
                cmd_contrib = CommandContribution(
                    id=id,
                    python_name=_python_name(_sample),
                    title=f"{key} sample",
                )
                self.contributions["commands"].append(cmd_contrib)
                s["command"] = id
            else:
                s["uri"] = str(_sample).replace(package_dir, r"${package}")

            self.contributions["sample_data"].append(s)

    def napari_experimental_provide_function(self, impl: HookImplementation):
        items: Union[Callable, List[Callable]] = impl.function()
        if not isinstance(items, list):
            items = [items]
        for idx, item in enumerate(items):
            try:

                cmd = f"{self.package}.{item.__name__}"
                py_name = _python_name(item)

                docsum = item.__doc__.splitlines()[0] if item.__doc__ else None
                cmd_contrib = CommandContribution(
                    id=cmd, python_name=py_name, title=docsum or item.__name__
                )
                self.contributions["commands"].append(cmd_contrib)

                wdg_contrib = WidgetContribution(
                    command=cmd,
                    display_name=item.__name__.replace("_", " "),
                    autogenerate=True,
                )
                self.contributions["widgets"].append(wdg_contrib)

            except Exception as e:
                msg = (
                    f"Error converting function [{idx}] "
                    f"from {impl.function.__module__!r}:\n{e}"
                )
                warnings.warn(msg)

    def napari_experimental_provide_dock_widget(self, impl: HookImplementation):
        WidgetCallable = Union[Callable, Tuple[Callable, dict]]
        items: Union[WidgetCallable, List[WidgetCallable]] = impl.function()
        if not isinstance(items, list):
            items = [items]  # pragma: no cover

        for idx, item in enumerate(items):
            if isinstance(item, tuple):
                wdg_creator = item[0]
                kwargs = item[1] if len(item) > 1 else {}
            else:
                wdg_creator, kwargs = (item, {})
            if not callable(wdg_creator) and isinstance(kwargs, dict):
                warnings.warn(f"Invalid widget spec: {wdg_creator}, {kwargs}")
                continue

            try:
                self._create_widget_contrib(impl, wdg_creator, kwargs)
            except Exception as e:
                msg = (
                    f"Error converting dock widget [{idx}] "
                    f"from {impl.function.__module__!r}:\n{e}"
                )
                warnings.warn(msg)

    def _create_widget_contrib(self, impl, wdg_creator, kwargs, is_function=False):
        # Get widget name
        func_name = getattr(wdg_creator, "__name__", "")
        wdg_name = str(kwargs.get("name", "")) or _camel_to_spaces(func_name)

        # in some cases, like partials and magic_factories, there might not be an
        # easily accessible python name (from __module__.__qualname__)...
        # so first we look for this object in the module namespace
        py_name = None
        cmd = None
        for local_name, val in impl.function.__globals__.items():
            if val is wdg_creator:
                py_name = f"{impl.function.__module__}:{local_name}"
                cmd = f"{self.package}.{local_name}"
                break
        else:
            try:
                py_name = _python_name(wdg_creator)
                cmd = (
                    f"{self.package}.{func_name or wdg_name.lower().replace(' ', '_')}"
                )
            except AttributeError:
                pass

        if not py_name:
            raise ValueError(
                "No suitable python name to point to. "
                "Is this a locally defined function or partial?"
            )

        # let these raise exceptions here immediately if they don't validate
        cmd_contrib = CommandContribution(
            id=cmd, python_name=py_name, title=f"Create {wdg_name}"
        )
        wdg_contrib = WidgetContribution(command=cmd, display_name=wdg_name)
        self.contributions["commands"].append(cmd_contrib)
        self.contributions["widgets"].append(wdg_contrib)

    def napari_get_writer(self, impl: HookImplementation):
        warnings.warn(
            "Found a multi-layer writer, but it's not convertable. "
            "Please add the writer manually."
        )
        return NotImplemented

    def napari_write_image(self, impl: HookImplementation):
        self._parse_writer(impl, "image")

    def napari_write_labels(self, impl: HookImplementation):
        self._parse_writer(impl, "labels")

    def napari_write_points(self, impl: HookImplementation):
        self._parse_writer(impl, "points")

    def napari_write_shapes(self, impl: HookImplementation):
        self._parse_writer(impl, "shapes")

    def napari_write_vectors(self, impl: HookImplementation):
        self._parse_writer(impl, "vectors")

    def _parse_writer(self, impl: HookImplementation, layer: str):
        id = self.add_command(impl)
        self.contributions["writers"].append(
            {
                "command": id,
                "layer_types": [layer],
                "display_name": layer,
                "filename_extensions": ["<EDIT_ME>"],
            }
        )

    def add_command(self, impl: HookImplementation, py_name: str = "") -> str:
        name = impl.specname.replace("napari_", "")
        id = f"{self.package}.{name}"
        title = " ".join(name.split("_")).title()
        if not py_name:
            py_name = _python_name(impl.function)
        c = CommandContribution(id=id, python_name=py_name, title=title)
        self.contributions["commands"].append(c)
        return id


def _safe_key(key: str) -> str:
    return (
        key.lower()
        .replace(" ", "_")
        .replace("-", "_")
        .replace("(", "")
        .replace(")", "")
        .replace("[", "")
        .replace("]", "")
    )


def _python_name(object):
    return f"{object.__module__}:{object.__qualname__}"


def _luma(r, g, b):
    # https://en.wikipedia.org/wiki/Luma_(video)
    return 0.2126 * r + 0.7152 * g + 0.0722 * b  # per ITU-R BT.709


_camel_to_spaces_pattern = re.compile(r"((?<=[a-z])[A-Z]|(?<!\A)[A-R,T-Z](?=[a-z]))")


def _camel_to_spaces(val):
    return _camel_to_spaces_pattern.sub(r" \1", val)


def get_top_module_path(package_name, top_module: Optional[str] = None) -> Path:
    dist = metadata.distribution(package_name)
    if not top_module:
        top_mods = (dist.read_text("top_level.txt") or "").strip().splitlines()
        if not top_mods:
            raise ValueError(
                "Could not detect a top level module in distribution metadata "
                f"for {package_name}"
            )
        top_module = top_mods[0]

    path = Path(dist.locate_file(top_module))
    assert path.is_dir()
    return path


def convert_repository(
    path: Union[Path, str], mf_name: str = "napari.yaml", dry_run=False
) -> Tuple[PluginManifest, Path]:
    """Convert repository at `path` to new npe2 style."""
    path = Path(path)

    # get the info we need and create a manifest
    info = get_package_dir_info(path)
    manifest = manifest_from_npe1(info.package_name)
    top_module = get_top_module_path(info.package_name, info.top_module)
    if not top_module.is_dir():
        raise ValueError(
            f"Detection of top-level module failed. {top_module} is not a directory."
        )
    mf_path = top_module / mf_name

    if dry_run:
        return manifest, mf_path

    # update the entry_points in setup.cfg/setup.py
    if info.setup_cfg:
        _write_new_setup_cfg_ep(info, mf_name)
    # or tell them to do it themselves in setup.py
    else:
        # tell them to do it manually
        warnings.warn(
            "\nCannot auto-update setup.py, please edit setup.py as follows:\n"
            "  1. remove the `napari.plugin` entry_point\n"
            "  2. add the following entry_point:"
            f"""
       entry_points={{
           "{NPE2_EP}": [
               "{info.package_name} = {info.top_module}:{mf_name}",
           ],
       }},
       package_data={{"{info.top_module}": ["{mf_name}"]}},
"""
        )

    # write the yaml to top_module/napari.yaml
    mf_path.write_text(manifest.yaml())
    return manifest, mf_path


def _write_new_setup_cfg_ep(info: PluginPackage, mf_name: str):
    assert info.setup_cfg
    p = ConfigParser(comment_prefixes="/", allow_no_value=True)  # preserve comments
    p.read(info.setup_cfg)
    mf_path = f"{info.top_module}:{mf_name}"
    new_ep = f"\n{info.package_name} = {mf_path}"
    if "options.entry_points" not in p.sections():
        p.add_section("options.entry_points")
    p.set("options.entry_points", NPE2_EP, new_ep)
    if "options.package_data" not in p.sections():
        p.add_section("options.package_data")
    p.set("options.package_data", info.top_module, mf_name)
    p.remove_option("options.entry_points", NPE1_EP)
    with open(info.setup_cfg, "w") as fh:
        p.write(fh)


def get_package_dir_info(path: Union[Path, str]) -> PluginPackage:
    """Attempts to *statically* get plugin info from a package directory."""
    path = Path(path).absolute()
    if not path.is_dir():
        raise ValueError(f"Provided path is not a directory: {path}")

    _name = None
    _entry_points: List[List[str]] = []
    _setup_cfg = None
    p = None

    # check for setup.cfg
    setup_cfg = path / "setup.cfg"
    if setup_cfg.exists():
        _setup_cfg = setup_cfg
        p = ConfigParser()
        p.read(setup_cfg)
        _name = p.get("metadata", "name", fallback=None)
        eps = p.get("options.entry_points", NPE1_EP, fallback="").strip()
        _entry_points = [[i.strip() for i in ep.split("=")] for ep in eps.splitlines()]

    if not _name or not _entry_points:
        # check for setup.py
        setup_py = path / "setup.py"
        if setup_py.exists():
            node = ast.parse(setup_py.read_text())
            visitor = _SetupVisitor()
            visitor.visit(node)
            _name = _name or visitor._name
            if visitor._entry_points and not _entry_points:
                _entry_points = visitor._entry_points
                _setup_cfg = None  # the ep metadata wasn't in setupcfg

    if _name and _entry_points:
        ep_name, ep_value = next(iter(_entry_points), ["", ""])
        top_mod = ep_value.split(".", 1)[0]
        return PluginPackage(_name, ep_name, ep_value, top_mod, _setup_cfg)

    msg = f'Could not detect first gen napari plugin package at "{path}".'
    if p is not None and p.get("options.entry_points", NPE2_EP, fallback=False):
        msg += f" Found a {NPE2_EP} entry_point. Is this package already converted?"
    raise ValueError(msg)


class _SetupVisitor(ast.NodeVisitor):
    """Visitor to statically determine metadata from setup.py"""

    def __init__(self) -> None:
        super().__init__()
        self._name: str = ""
        self._entry_points: List[List[str]] = []  # [[name, value], ...]

    def visit_Call(self, node: ast.Call) -> Any:
        if getattr(node.func, "id", "") != "setup":
            return
        for kw in node.keywords:
            if kw.arg == "name":
                self._name = getattr(kw.value, "id", "")
            if kw.arg == "entry_points":
                eps: dict = ast.literal_eval(kw.value)
                for k, v in eps.items():
                    if k != NPE1_EP:
                        continue
                    for item in v:
                        self._entry_points.append([i.strip() for i in item.split("=")])



from __future__ import annotations

__all__ = ["PluginContext", "PluginManager"]

import os
from collections import Counter
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    DefaultDict,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from intervaltree import IntervalTree

from ._command_registry import CommandRegistry
from .manifest import PluginManifest
from .manifest.io import LayerType

if TYPE_CHECKING:
    from .manifest.commands import CommandContribution
    from .manifest.contributions import ContributionPoints
    from .manifest.io import ReaderContribution, WriterContribution
    from .manifest.menus import MenuItem
    from .manifest.sample_data import SampleDataContribution
    from .manifest.submenu import SubmenuContribution
    from .manifest.themes import ThemeContribution
    from .manifest.widgets import WidgetContribution

    T = TypeVar("T")

    class Interval(tuple, Generic[T]):
        begin: int
        end: int
        data: T

    class TypedIntervalTree(IntervalTree, Generic[T]):
        def addi(self, begin: int, end: int, data: T) -> None:
            ...

        def __getitem__(self, index: Union[int, slice]) -> Set[Interval[T]]:
            ...


PluginName = str  # this is `PluginManifest.name`


class _ContributionsIndex:
    def __init__(self) -> None:
        self._submenus: Dict[str, SubmenuContribution] = {}
        self._commands: Dict[str, Tuple[CommandContribution, PluginName]] = {}
        self._themes: Dict[str, ThemeContribution] = {}
        self._widgets: List[WidgetContribution] = []
        self._readers: DefaultDict[str, List[ReaderContribution]] = DefaultDict(list)
        self._samples: DefaultDict[str, List[SampleDataContribution]] = DefaultDict(
            list
        )
        self._writers_by_type: DefaultDict[
            LayerType, TypedIntervalTree[WriterContribution]
        ] = DefaultDict(IntervalTree)
        self._writers_by_command: DefaultDict[
            str, List[WriterContribution]
        ] = DefaultDict(list)

    def index_contributions(self, ctrb: ContributionPoints, key: PluginName):
        if ctrb.sample_data:
            self._samples[key] = ctrb.sample_data
        for cmd in ctrb.commands or []:
            self._commands[cmd.id] = cmd, key
        for subm in ctrb.submenus or []:
            self._submenus[subm.id] = subm
        for theme in ctrb.themes or []:
            self._themes[theme.id] = theme
        self._widgets.extend(ctrb.widgets or [])
        for reader in ctrb.readers or []:
            for pattern in reader.filename_patterns:
                self._readers[pattern].append(reader)
            if reader.accepts_directories:
                self._readers[""].append(reader)
        for writer in ctrb.writers or []:
            self._writers_by_command[writer.command].append(writer)
            for c in writer.layer_type_constraints():
                self._writers_by_type[c.layer_type].addi(*c.bounds, writer)


class PluginManager:
    __instance: Optional[PluginManager] = None  # a global instance
    _contrib: _ContributionsIndex

    def __init__(self, reg: Optional[CommandRegistry] = None) -> None:
        self._command_registry = reg or CommandRegistry()
        self._contexts: Dict[PluginName, PluginContext] = {}
        self._manifests: Dict[PluginName, PluginManifest] = {}
        self.discover()  # TODO: should we be immediately discovering?

    @property
    def commands(self) -> CommandRegistry:
        return self._command_registry

    def discover(self, paths: Sequence[str] = ()) -> None:
        """Discover and index plugin manifests in the environment.

        Parameters
        ----------
        paths : Sequence[str], optional
            Optional list of strings to insert at front of sys.path when discovering.
        """
        self._contrib = _ContributionsIndex()
        self._manifests.clear()

        for result in PluginManifest.discover(paths=paths):
            if result.manifest is None:
                continue
            mf = result.manifest
            if mf.name in self._manifests:
                continue
            self._manifests[mf.name] = mf
            if mf.contributions:
                self._contrib.index_contributions(mf.contributions, mf.name)

    def get_manifest(self, key: str) -> PluginManifest:
        key = str(key).split(".")[0]
        try:
            return self._manifests[key]
        except KeyError:
            raise KeyError(f"Manifest key {key!r} not found in {list(self._manifests)}")

    def get_command(self, command_id: str) -> CommandContribution:
        return self._contrib._commands[command_id][0]

    def get_submenu(self, submenu_id: str) -> SubmenuContribution:
        return self._contrib._submenus[submenu_id]

    def iter_menu(self, menu_key: str) -> Iterator[MenuItem]:
        for mf in self._manifests.values():
            if mf.contributions:
                yield from getattr(mf.contributions.menus, menu_key, [])

    def iter_themes(self) -> Iterator[ThemeContribution]:
        yield from self._contrib._themes.values()

    def activate(self, key: PluginName) -> PluginContext:
        """Activate plugin with `key`.

        This does the following:
            - finds the manifest for the associated plugin key
            - gets or creates a PluginContext for the plugin
            - bails if it's already activated
            - otherwise calls the plugin's activate() function, passing the Context.
            - imports any commands that were declared as python_name:
        """
        # TODO: this is an important function... should be carefully considered
        try:
            mf = self._manifests[key]
        except KeyError:
            raise KeyError(f"Cannot activate unrecognized plugin key {key!r}")

        # create the context that will be with this plugin for its lifetime.
        ctx = self.get_context(key)
        if ctx._activated:
            # prevent "reactivation"
            return ctx

        try:
            mf._call_func_in_plugin_entrypoint("activate", args=(ctx,))
            ctx._activated = True
        except Exception as e:  # pragma: no cover
            self._contexts.pop(key, None)
            raise type(e)(f"Activating plugin {key!r} failed: {e}") from e

        if mf.contributions and mf.contributions.commands:
            for cmd in mf.contributions.commands:
                if cmd.python_name and cmd.id not in self.commands:
                    self.commands.register(cmd.id, cmd.python_name)

        return ctx

    def deactivate(self, key: PluginName) -> None:
        if key not in self._contexts:
            return
        mf = self._manifests[key]
        ctx = self._contexts.pop(key)
        mf._call_func_in_plugin_entrypoint("deactivate", args=(ctx,))
        ctx._dispose()

    def iter_compatible_readers(
        self, path: Union[str, Path]
    ) -> Iterator[ReaderContribution]:
        from fnmatch import fnmatch

        if isinstance(path, list):
            return NotImplemented
        if os.path.isdir(path):
            yield from self._contrib._readers[""]
        else:
            seen: Set[str] = set()
            for ext, readers in self._contrib._readers.items():
                if ext and fnmatch(str(path), ext):
                    for r in readers:
                        if r.command in seen:
                            continue
                        seen.add(r.command)
                        yield r

    @classmethod
    def instance(cls) -> PluginManager:
        """Singleton instance."""
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    def get_context(self, plugin_key: PluginName) -> PluginContext:
        if plugin_key not in self._contexts:
            self._contexts[plugin_key] = PluginContext(plugin_key, reg=self.commands)
        return self._contexts[plugin_key]

    def iter_sample_data(self) -> Iterator[Tuple[str, List[SampleDataContribution]]]:
        """Iterates over (plugin_name, [sample_contribs])."""
        yield from self._contrib._samples.items()

    def get_writer_for_command(self, command: str) -> Optional[WriterContribution]:
        writers = self._contrib._writers_by_command[command]
        return writers[0] if writers else None

    def iter_widgets(self) -> Iterator[WidgetContribution]:
        yield from self._contrib._widgets

    def iter_compatible_writers(
        self, layer_types: Sequence[str]
    ) -> Iterator[WriterContribution]:
        """Attempt to match writers that consume all layers."""

        if not layer_types:
            return

        # First count how many of each distinct type are requested. We'll use
        # this to get candidate writers compatible with the requested count.
        counts = Counter(layer_types)

        def _get_candidates(lt: LayerType) -> Set[WriterContribution]:
            return {
                v.data for v in self._contrib._writers_by_type[lt][counts[lt]] or []
            }

        types = iter(LayerType)
        candidates = _get_candidates(next(types))
        for lt in types:
            if candidates:
                candidates &= _get_candidates(lt)
            else:
                break

        def _writer_key(writer: WriterContribution) -> Tuple[bool, int, int, List[str]]:
            # 1. writers with no file extensions (like directory writers) go last
            no_ext = len(writer.filename_extensions) == 0

            # 2. more "specific" writers first
            nbounds = sum(not c.is_zero() for c in writer.layer_type_constraints())

            # 3. then sort by the number of listed extensions
            #    (empty set of extensions goes last)
            ext_len = len(writer.filename_extensions)

            # 4. finally group related extensions together
            exts = writer.filename_extensions
            return (no_ext, nbounds, ext_len, exts)

        yield from sorted(
            candidates,
            key=_writer_key,
        )

    def get_writer(
        self, path: str, layer_types: Sequence[str], plugin_name: Optional[str] = None
    ) -> Tuple[Optional[WriterContribution], str]:
        """Get Writer contribution appropriate for `path`, and `layer_types`.

        When `path` has a file extension, find a compatible writer that has
        that same extension. When there is no extension and only a single layer,
        find a compatible writer and append the extension.
        Otherwise, find a compatible no-extension writer and write to that.
        No-extension writers typically write to a folder.

        Parameters
        ----------
        path : str
            Path to write
        layer_types : Sequence[str]
            Sequence of layer type strings (e.g. ['image', 'labels'])
        plugin_name : Optional[str], optional
            Optional name of plugin to use. by default None (look for plugin)

        Returns
        -------
        Tuple[Optional[WriterContribution], str]
            WriterContribution and path that will be written.
        """
        ext = Path(path).suffix.lower() if path else ""

        for writer in self.iter_compatible_writers(layer_types):
            if plugin_name and not writer.command.startswith(plugin_name):
                continue

            if (
                ext
                and ext in writer.filename_extensions
                or not ext
                and len(layer_types) != 1
                and not writer.filename_extensions
            ):
                return writer, path
            elif not ext and len(layer_types) == 1:  # No extension, single layer.
                ext = next(iter(writer.filename_extensions), "")
                return writer, path + ext
            # When the list of extensions for the writer doesn't match the
            # extension in the filename, keep searching.

        # Nothing got found
        return None, path


class PluginContext:
    """An object that can contain information for a plugin over its lifetime."""

    # stores all created contexts (currently cleared by `PluginManager.deactivate`)

    def __init__(
        self, plugin_key: PluginName, reg: Optional[CommandRegistry] = None
    ) -> None:
        self._activated = False
        self.plugin_key = plugin_key
        self._command_registry = reg or PluginManager.instance().commands
        self._imports: Set[str] = set()  # modules that were imported by this plugin
        self._disposables: Set[Callable] = set()  # functions to call when deactivating

    def _dispose(self):
        for dispose in self._disposables:
            dispose()

    def register_command(self, id: str, command: Optional[Callable] = None):
        def _inner(command):
            self._disposables.add(self._command_registry.register(id, command))
            return command

        return _inner if command is None else _inner(command)



import pytest

from npe2._command_registry import CommandHandler, CommandRegistry
from npe2._plugin_manager import PluginManager


def test_plugin_manager(sample_path):
    pm = PluginManager()
    pm.discover()
    assert len(pm._manifests) == 0
    pm.discover([sample_path])
    assert len(pm._manifests) == 1
    assert pm.get_command("my_plugin.hello_world")

    assert "my_plugin" not in pm._contexts
    ctx = pm.activate("my_plugin")
    assert "my_plugin" in pm._contexts

    # dual activation is prevented
    assert pm.activate("my_plugin") is ctx

    with pytest.raises(KeyError):
        pm.activate("not a thing")

    assert pm.get_command("my_plugin.hello_world")
    with pytest.raises(KeyError):
        pm.get_command("my_plugin.not_a_thing")

    assert pm.get_submenu("mysubmenu")
    assert len(list(pm.iter_menu("/napari/layer_context"))) == 2

    # deactivation
    assert "my_plugin" in pm._contexts
    pm.deactivate("my_plugin")
    assert "my_plugin" not in pm._contexts
    pm.deactivate("my_plugin")  # second time is a no-op
    assert "my_plugin" not in pm._contexts


def test_command_handler():
    with pytest.raises(RuntimeError):
        # cannot resolve something without either a python_name or function
        CommandHandler("hi").resolve()

    with pytest.raises(RuntimeError):
        # cannot resolve something without either a python_name or function
        CommandHandler("hi", python_name="cannot.import.this").resolve()


def test_command_reg_register():
    reg = CommandRegistry()
    with pytest.raises(ValueError):
        # must register non empty string id
        reg.register(1, lambda: None)  # type: ignore
    with pytest.raises(TypeError):
        # neither a string or a callable
        reg.register("other.id", 8)  # type: ignore

    with pytest.raises(ValueError):
        # must register non empty string id
        reg.register("some.id", "1_is_not.a_valid_python_name")
    reg.register("some.id", "this.is.a_valid_python_name")

    with pytest.raises(ValueError):
        # already registered
        reg.register("some.id", "this.is.a_valid_python_name")


def test_command_reg_get():
    def f(x, y):
        return x + y

    reg = CommandRegistry()
    reg.register("id", f)
    assert "id" in reg
    assert reg.get("id") is f
    assert reg.execute("id", (1, 2)) == 3



from typing import List, Optional, Union

from pydantic import BaseModel, Field, ValidationError, root_validator

from .utils import Executable

# # napari provides these
# class Menu(BaseModel):
#     key: str
#     id: int
#     description: str
#     supports_submenus: bool = True
#     deprecation_message: Optional[str]


# napari_menus = [
#     Menu(key="command_pallete", id=0, description="The Command Palette"),
#     Menu(
#         key="layers__context", id=1, description="The layer list context menu"
#     ),
#     Menu(
#         key="layers__context", id=1, description="The layer list context menu"
#     ),
# ]


# user provides this
class _MenuItem(BaseModel):
    when: Optional[str] = Field(
        description="Condition which must be true to show this item"
    )
    group: Optional[str] = Field(description="Group into which this item belongs")


class Submenu(_MenuItem):
    submenu: str = Field(
        ...,
        description="Identifier of the submenu to display in this item."
        "The submenu must be declared in the 'submenus' -section",
    )
    # if submenu doesn't exist, you get:
    # Menu item references a submenu ...` which is not defined in the 'submenus' section


class MenuCommand(_MenuItem, Executable):
    command: str = Field(
        ...,
        description="Identifier of the command to execute. "
        "The command must be declared in the 'commands' section",
    )
    # if command doesn't exist, you get:
    # "Menu item references a command `...` which is not defined in the
    # 'commands' section."
    alt: Optional[str] = Field(
        description="Identifier of an alternative command to execute. "
        "It will be shown and invoked when pressing Alt while opening a menu."
        "The command must be declared in the 'commands' section"
    )
    # if command doesn't exist, you get:
    # "Menu item references an alt-command  `...` which is not defined in
    # the 'commands' section."


MenuItem = Union[MenuCommand, Submenu]


# how to do something like layers/context
class MenusContribution(BaseModel):
    # TODO: list of (str, menu item) coerce to dict/MenuItem
    # TODO: define convention around strings
    command_pallete: Optional[List[MenuItem]]
    layers__context: Optional[List[MenuItem]]
    plugins__widgets: Optional[List[MenuItem]]
    test_menu: Optional[List[MenuItem]]

    class Config:
        extra = "allow"

    @root_validator
    def _validate_extra(cls, values: dict):
        """Plugins may declare custom menu IDs... we make sure they are valid here.

        They become accessible as attributes on the MenusContribution instance.
        """

        # get validator... all of these fields have the same type
        # (Optional[List[MenuItem]])
        validate = list(MenusContribution.__fields__.values())[0].validate

        for i, (key, val) in enumerate(values.items()):
            if key not in cls.__fields__:
                val, err = validate(val, {}, loc=str(i))
                if err:
                    raise ValidationError([err], cls)
                values[key] = val
        return values
